import io
import math
import typing as t
from multiprocessing import Pool
from pathlib import Path

import cloudpickle as cp
import httpx

from pipeline.objects import Graph, PipelineFile
from pipeline.util.logging import _print
from pipeline.v3 import http
from pipeline.v3.schemas.runs import Run


def upload_pipeline(
    graph: Graph,
    environment_id_or_name: t.Union[str, int],
):
    if graph._has_run_startup:
        raise Exception("Graph has already been run, cannot upload")

    for variable in graph.variables:
        if isinstance(variable, PipelineFile):
            variable_path = Path(variable.path)
            if not variable_path.exists():
                raise FileNotFoundError(
                    f"File not found for variable (path={variable.path}, "
                    f"variable={variable.name})"
                )

            variable_file = variable_path.open("rb")
            try:
                res = http.post_files(
                    "/v3/pipeline_files",
                    files=dict(pfile=variable_file),
                    progress=True,
                )

                new_path = res.json()["path"]
                variable.path = new_path
                variable.remote_id = res.json()["id"]
            finally:
                variable_file.close()

    with open("graph.tmp", "wb") as tmp:
        tmp.write(cp.dumps(graph))

    graph_file = open("graph.tmp", "rb")

    params = dict()
    if graph.min_gpu_vram_mb is not None:
        params["gpu_memory_min"] = graph.min_gpu_vram_mb

    if isinstance(environment_id_or_name, int):
        params["environment_id"] = environment_id_or_name
    elif isinstance(environment_id_or_name, str):
        params["environment_name"] = environment_id_or_name

    res = http.post_files(
        "/v3/pipelines",
        files=dict(graph=graph_file),
        params=params,
    )

    graph_file.close()

    return res


def run_pipeline(
    graph_id: str,
    data: t.Any,
    *,
    return_response: bool = False,
    async_run: bool = False,
) -> t.Union[Run, httpx.Response]:
    data_obj = io.BytesIO(cp.dumps(data))

    res = http.post_files(
        "/v3/runs",
        params=dict(
            graph_id=graph_id,
            async_run=async_run,
        ),
        files=dict(input_data=data_obj),
    )

    if return_response:
        return res

    if res.status_code == 500:
        _print(
            f"Failed run (status={res.status_code}, text={res.text}, "
            f"headers={res.headers})",
            level="ERROR",
        )
        raise Exception(f"Error: {res.status_code}, {res.text}", res.status_code)
    elif res.status_code == 429:
        raise Exception(
            "Too many requests, please try again later",
            res.status_code,
        )

    try:
        res.raise_for_status()
    except httpx.HTTPStatusError as exc:
        raise Exception(f"HTTP error: {exc.response.status_code}, {exc.response.text}")

    # Everything is okay!
    run_get = Run.parse_obj(res.json())

    return run_get


def get_pipeline_run(run_id: int) -> Run:
    http_res = http.get(f"/v3/runs/{run_id}")

    run_get = Run.parse_obj(http_res.json())
    return run_get


def map_pipeline_mp(array: list, graph_id: str, *, pool_size=8):
    results = []

    num_batches = math.ceil(len(array) / pool_size)
    for b in range(num_batches):
        start_index = b * pool_size
        end_index = min(len(array), start_index + pool_size)

        inputs = [[graph_id, item] for item in array[start_index:end_index]]

        with Pool(pool_size) as p:
            batch_res = p.starmap(run_pipeline, inputs)
            results.extend(batch_res)

    return results
